# -*- coding: utf-8 -*-
"""Provides decorator classes for assorted functions"""

from logging import MythLog
from exceptions import MythDBError

from cStringIO import StringIO
from select import select
from time import time
import socket

class schemaUpdate( object ):
    # decorator class for database updates
    # TODO: do locking and lock checking
    #       if interactive terminal, ask for update permission
    #       perform database backup (partial?)
    def __init__(self, func):
        self.func = func
        self.__doc__ = self.func.__doc__
        self.__name__ = self.func.__name__
        self.__module__ = self.func.__module__

        self.schemavar = func(-1)

    def __call__(self, db):
        log = MythLog('Schema Update')

        while True:
            try:
                updates, newver = self.func(db.settings.NULL[self.schemavar])
            except StopIteration:
                break

            log(log.IMPORTANT, 'Updating %s from %s to %s' % \
                        (schemaname, db.settings.NULL[self.schemavar], newver))
            c = db.cursor()

            try:
                for sql, values in updates:
                    c.execute(sql, values)
            except Exception, e:
                log(log.IMPORTANT, 'Update of %s failed' % self.schemavar)
                raise MythDBError(MythError.DB_SCHEMAUPDATE, e.args)

            c.close()
            db.settings.NULL[self.schemavar] = newver

class databaseSearch( object ):
    # decorator class for database searches
    def __init__(self, func):
        # set function and update strings
        self.func = func
        self.__doc__ = self.func.__doc__
        self.__name__ = self.func.__name__
        self.__module__ = self.func.__module__

        # process joins
        res = list(self.func(self, True))
        self.table = res[0]
        self.dbclass = res[1]
        self.require = res[2]
        if len(res) > 3:
            self.joins = res[3:]
        else:
            self.joins = ()

    def __get__(self, inst, own):
        # set instance and return self
        self.inst = inst
        return self

    def __call__(self, **kwargs):
        where = []
        fields = []
        joinbit = 0

        # loop through inputs
        for key, val in kwargs.items():
            if val is None:
                continue
            # process custom query
            if key == 'custom':
                custwhere = {}
                custwhere.update(val)
                for k,v in custwhere.items():
                    where.append(k)
                    fields.append(v)
                continue
            # let function process remaining queries
            res = self.func(self.inst, key=key, value=val)
            errstr = "%s got an unexpected keyword argument '%s'"
            if res is None:
                if 'not' not in key:
                    raise TypeError(errstr % (self.__name__, key))
                res = list(self.func(self.inst, key=key[3:], value=val))
                if res is None:
                    raise TypeError(errstr % (self.__name__, key))
                res[0] = 'NOT '+res[0]

            if len(res) == 3:
                where.append(res[0])
                fields.append(res[1])
                joinbit = joinbit|res[2]
            elif len(res) == 4:
                lval = val.split(',')
                where.append('(%s)=%d' %\
                    (self.buildQuery(
                        (   self.buildJoinOn(res[3]),
                            '(%s)' % \
                                 ' OR '.join(['%s=%%s' % res[0] \
                                                    for f in lval])),
                        'COUNT( DISTINCT %s )' % res[0],
                        res[1],
                        res[2]),
                    len(lval)))
                fields += lval
                            
        for key in self.require:
            if key not in kwargs:
                res = self.func(self.inst, key=key)
                if res is None:
                    continue
                where.append(res[0])
                fields.append(res[1])
                joinbit = joinbit|res[2]

        # process query
        query = self.buildQuery(where, joinbit=joinbit)
        c = self.inst.cursor(self.inst.log)
        if len(where) > 0:
            c.execute(query, fields)
        else:
            c.execute(query)

        row = c.fetchone()
        while row is not None:
            yield self.dbclass.fromRaw(row, self.inst)
            row = c.fetchone()

    def buildJoinOn(self, i):
        if len(self.joins[i]) == 3:
            # shared field name
            on = ' AND '.join(['%s.%s=%s.%s' % \
                    (self.joins[i][0], field,\
                     self.joins[i][1], field)\
                             for field in self.joins[i][2]])
        else:
            # separate field name
            on = ' AND '.join(['%s.%s=%s.%s' % \
                    (self.joins[i][0], ffield,\
                     self.joins[i][1], tfield)\
                             for ffield, tfield in zip(\
                                    self.joins[i][2],\
                                    self.joins[i][3])])
        return on

    def buildQuery(self, where, select=None, tfrom=None, joinbit=0):
        sql = 'SELECT '
        if select:
            sql += select
        elif tfrom:
            sql += tfrom+'.*'
        else:
            sql += self.table+'.*'

        sql += ' FROM '
        if tfrom:
            sql += tfrom
        else:
            sql += self.table

        if joinbit:
            for i in range(len(self.joins)):
                if (2**i)&joinbit:
                    sql += ' JOIN %s ON %s' % \
                            (self.joins[i][0], self.buildJoinOn(i))

        if len(where):
            sql += ' WHERE '
            sql += ' AND '.join(where)

        return sql

class SplitInt( object ):
    @staticmethod
    def joinInt(high,low):
        """obj.joinInt(high, low) -> 64-bit int, from two signed ints"""
        return (high + (low<0))*2**32 + low

    @staticmethod
    def splitInt(integer):
        """obj.joinInt(64-bit int) -> (high, low)"""
        return integer/(2**32),integer%2**32 - (integer%2**32 > 2**31)*2**32

class CMPVideo( object ):
    def __cmp__(self, other):
        res = cmp(self.title, other.title)
        if res:
            return res
        return cmp(self.subtitle, other.subtitle)

class CMPRecord( CMPVideo ):
    def __cmp__(self, other):
        if ('starttime' not in other) or ('video' not in other):
            return CMPVideo.__cmd__(self, other)
        res = cmp(self.starttime, other.starttime)
        if res:
            return res
        return cmp(self.chanid, other.chanid)

class deadlinesocket( socket.socket ):
    def __init__(self, *args, **kwargs):
        socket.socket.__init__(self, *args, **kwargs)
        self.log = MythLog('Python Socket')
        self.setdeadline(10.0)

    def connect(self, *args, **kwargs):
        socket.socket.connect(self, *args, **kwargs)
        self.setblocking(0)

    def getdeadline(self): return self._deadline
    def setdeadline(self, deadline): self._deadline = deadline

    def dlrecv(self, bufsize, flags=0, deadline=None):
        t = time()
        if deadline is None:
            deadline = t + self._deadline

        buff = StringIO()
        # loop until necessary data has been received
        while bufsize > buff.tell():
            # wait for data on the socket
            if len(select([self],[],[], deadline-t)[0]) == 0:
                # deadline reached, terminate
                return u''
            # append response to buffer
            buff.write(self.recv(bufsize-buff.tell(), flags))
            if t >= deadline:
                break
            t = time()
        return buff.getvalue()

    def dlexpect(self, pattern, flags=0, deadline=None):
        t = time()
        if deadline is None:
            deadline = t + self._deadline

        buff = StringIO()
        # loop until pattern has been found
        while not pattern.search(buff.getvalue()):
            if len(select([self],[],[], deadline-t)[0]) == 0:
                return ''
            buff.write(self.recv(100, flags))
            if t >= deadline:
                break
            t = time()
        return buff.getvalue()

    def recvheader(self, flags=0, deadline=None):
        try:
            size = int(self.dlrecv(8, flags, deadline))
        except:
            return ''
        data = self.dlrecv(size, flags, deadline)
        self.log(MythLog.SOCKET|MythLog.NETWORK, \
                            'read <-- %d' % size, data)
        return data

    def sendheader(self, data, flags=0):
        self.log(MythLog.SOCKET|MythLog.NETWORK, \
                            'write --> %d' % len(data), data)
        data = '%-8d%s' % (len(data), data)
        self.send(data, flags)
