#!/usr/bin/env python

from MythTV import *
from time import mktime
from datetime import datetime, timedelta
from socket import gethostname
import traceback
import weakref
import sys
import os

MythLog._setlevel('none')

DBCONN = {'DBHostName': 'sql',
          'DBPort': 3306,
          'DBName': 'mythconverg_trunk',
          'DBUserName': 'mythtv',
          'DBPassword': 'mythtv'}
BEIP = '127.0.0.1'
FRONTEND = ['127.0.0.1',6546]
ERRORLIST = []
OPTS = {}

class testgroup( object ):
    _current = []
    def __repr__(self):
        return '%s(%s)' % (self.name, self._children)
    def __init__(self, name=None):
        self._parent = self._current
        testgroup._current = self
        self._children = []
        self.name = name
        self._parent.append(self)
    def __call__(self):
        if self.name:
            pi('running %s' % self.name, newline=True, color='cyan')
            pi.raiseindent()
        for func in self._children: func()
        if self.name:
            pi.lowerindent()
    def append(self, func):
        self._children.append(func)
    def close(self):
        testgroup._current = self._parent
    @classmethod
    def getcurrent(cls):
        return cls._current

class Critical( Exception ): pass
class Skipped( Exception ): pass

class decor( object ):
    def __init__(self, func):
        self.func = func
        self.__doc__ = self.func.__doc__
        self.__name__ = self.func.__name__
        self.__module__ = self.func.__module__
    def __call__(self, *args, **kwargs):
        self.func(*args, **kwargs)

class required( decor ):
    def __call__(self, *args, **kwargs):
        try: self.func(*args, **kwargs)
        except KeyboardInterrupt: raise
        except: raise Critical

class requires( decor ):
    def __call__(self, *args, **kwargs):
        if not OPTS[self.name]:
            raise Skipped
        self.func(*args, **kwargs)

class requires_msearch( requires ): name = 'msearch'
class requires_dbconn( requires ): name = 'dbconn'
class requires_beconn( requires ): name = 'beconn'
class requires_xmlconn( requires ): name = 'xmlconn'
class requires_feconn( requires ): name = 'feconn'
class requires_dbcache( requires ): name = 'dbcache_normal'
class requires_becache( requires ): name = 'becache'
class requires_record_getall( requires ): name = 'record_getall'

class test( decor ):
    def __repr__(self):
        return self.__name__
    def __init__(self, func):
        decor.__init__(self, func)
        testgroup.getcurrent().append(self)
    def __call__(self, *args, **kwargs):
        res = None
        pi('running %s' % self.__name__[5:])
        try:
            res = self.func(*args, **kwargs)
            pi('PASSED', ralign=True, color='green')
            OPTS[self.__name__[5:]] = True
        except Skipped:
            pi('SKIPPED', ralign=True, color='yellow')
            OPTS[self.__name__[5:]] = False
        except Critical:
            pi('CRITICAL', ralign=True, color='magenta')
            ERRORLIST.append((self.__name__, traceback.format_exc()))
            raise
        except KeyboardInterrupt:
            raise
        except:
            pi('FAILED', ralign=True, color='red')
            ERRORLIST.append((self.__name__, traceback.format_exc()))
            OPTS[self.__name__[5:]] = False

class print_indent():
    width = 60
    colors = dict(zip(['black','red','green','yellow','blue','magenta','cyan','white'], range(8)))
    def __init__(self):
        self.col = 0
        self.indent = 0
    def __call__(self, text, newline=False, ralign=False, color=None):
        if self.col == 0:
            sys.stdout.write(' '*self.indent)
            self.col += self.indent
        size = len(text)
        if color is not None:
            text = '\033[3%dm%s\033[37m' % (self.colors[color], text)
        if ralign:
            newline = True
            sys.stdout.write(' '*(self.width-(self.col+size)))
        sys.stdout.write(text)
        self.col += size
        if newline:
            sys.stdout.write('\n')
            self.col = 0
    def raiseindent(self): self.indent += 4
    def lowerindent(self): self.indent = max(self.indent-4, 0)

runtests = testgroup()
#-----------------------------------------
# base classes
test_base = testgroup('base classes')

@test
@required
def test_orddict():
    od = OrdDict((('somedata',1),('otherdata',2)))
    #test attribute access
    od.somedata
    #test attribute set
    od.somedata = 2
    if not od.somedata == 2: raise Exception
    #test item set
    od['somedata'] = 3
    if not od['somedata'] == 3: raise Exception
    #test new attribute set
    od.moredata = 4
    if not od.moredata == 4: raise Exception
    #test iterators
    if ['somedata','otherdata','moredata'] != od.keys(): raise Exception(od.keys())
    if [3,2,4] != od.values(): raise Exception(od.values())
    #test copying
    od2 = od.copy()
    if not od2.moredata == 4: raise Exception

@test
def test_dictinvert():
    di,id = DictInvert.createPair({'somedata':1, 'otherdata':2})
    #test item set
    di['somedata'] = 3
    if not id[3] == 'somedata': raise Exception
    #test new item set
    id[4] = 'moredata'
    if not di['moredata'] == 4: raise Exception
    #test item delete
    del id[4]
    if 'moredata' in di: raise Exception

test_base.close()

#-----------------------------------------
# connections
test_connections = testgroup('connections')
@test
def test_msearch():
    ms = MSearch()
    try:
        ms.search().next()
    except StopIteration:
        raise Skipped

test_base_connections = testgroup('base connections')

@test
def test_dbconn():
    db = DBConnection(DBCONN)
    c = db.cursor()
    c.execute("""SELECT 1""")
    if not c.fetchone()[0] == 1: raise Exception

@test
def test_beconn(): be = BEConnection(BEIP, 6543)

@test
def test_xmlconn():
    xml = XMLConnection(BEIP, port=6544)
    xml._query()

@test
def test_feconn():
    fe = FEConnection(FRONTEND[0], FRONTEND[1])

test_base_connections.close()
test_connection_caches = testgroup('connection caches')
test_dbcache = testgroup('DBCache')

@test
@requires_dbconn
def test_dbcache_normal(): 
    global DBCACHE
    DBCACHE = DBCache()

@test
@requires_dbconn
def test_dbcache_manual_args():
    move_dbfile()
    try: DBCache(args=DBCONN)
    except: raise
    finally: move_dbfile(True)

@test
@requires_dbconn
def test_dbcache_manual_kw():
    move_dbfile()
    try:
        DBCache(DBHostName=DBCONN['DBHostName'], DBName=DBCONN['DBName'],
                DBUserName=DBCONN['DBUserName'], DBPassword=DBCONN['DBPassword'])
    except: raise
    finally: move_dbfile(True)

@test
@requires_dbconn
def test_dbcache_upnp():
    raise Skipped
    move_dbfile()
    try: DBCache()
    except: raise
    finally: move_dbfile(True)

@test
@requires_dbconn
def test_dbcache_recurse():
    db = DBCache(DBCACHE)

@test
@requires_dbcache
def test_dbcache_storagegroup():
    db = DBCACHE
    db.getStorageGroup().next()
    db.getStorageGroup(groupname='Default', 
                       hostname=gethostname()).next()

test_dbcache.close()

@test
@requires_dbcache
def test_becache():
    be = BECache(db=DBCACHE)
    host = be.host
    hostname = be.hostname
    BECache(host)
    BECache(hostname)

@test
@requires_feconn
def test_feconn_upnp():
    fe = FEConnection.fromUPNP().next()

test_connection_caches.close()
test_connections.close()

test_collections = testgroup('collections')
test_data_objects = testgroup('database objects')

test_record = testgroup('Record')

@test
def test_record_getall():
    recs = list(Record.getAllEntries(DBCACHE))

@test
@requires_record_getall
def test_record_openone():
    try:
        rec = Record.getAllEntries(DBCACHE).next()
    except StopIteration:
        raise Skipped
    rec2 = Record(rec.recordid, db=DBCACHE)

test_record.close()

test_recorded = testgroup('Recorded')
@test
def test_recorded_getall():
    rec = list(Recorded.getAllEntries(DBCACHE))

@test
def test_recorded_getone():
    rec = Recorded.getAllEntries(DBCACHE).next()
    rec2 = Recorded((rec.chanid, rec.starttime))

@test
def test_recorded_cast():
    rec = Recorded.getAllEntries(DBCACHE).next()
    list(rec.cast)
    member = rec.cast.SubData(zip(rec.cast._datfields, ['Test Actor', u'actor']))
    # test adding
    rec.cast.add(*member.values())
    rec.cast.commit()
    rec2 = Recorded((rec.chanid, rec.starttime), db=DBCACHE)
    if member not in rec2.cast:
        raise Exception('Cast data not added to database')

    # test deletion
    rec2.cast.delete(*member.values())
    rec2.cast.commit()
    rec2 = Recorded((rec.chanid, rec.starttime), db=DBCACHE)
    if member in rec2.cast:
        raise Exception('Cast data not removed from database')

@test
def test_recorded_seek():
    rec = Recorded.getAllEntries(DBCACHE).next()
    list(rec.seek)
    if len(rec.seek) == 0:
        raise Skipped
    mark = rec.seek[0].mark
    rec.seek[0].mark = mark+1
    rec.seek.commit()
    rec2 = Recorded((rec.chanid, rec.starttime), db=DBCACHE)
    if rec2.seek[0].mark != mark+1:
        raise Exception('Mark could not be updated')
    rec2.seek[0].mark = mark
    rec2.seek.commit()

@test
def test_recorded_markup():
    rec = Recorded.getAllEntries(DBCACHE).next()
    list(rec.markup)
    if len(rec.markup) == 0:
        raise Skipped
    mark = rec.markup[0].mark
    rec.markup[0].mark = mark+1
    rec.markup.commit()
    rec2 = Recorded((rec.chanid, rec.starttime), db=DBCACHE)
    if rec2.markup[0].mark != mark+1:
        raise Exception('Mark could not be updated')
    rec2.markup[0].mark = mark
    rec2.markup.commit()

test_recorded.close()

test_recordedprogram = testgroup('RecordedProgram')

@test
@requires_becache
def test_recordedprogram_getall():
    rec = list(RecordedProgram.getAllEntries(DBCACHE))

@test
@requires_becache
def test_recordedprogram_getone():
    rec = RecordedProgram.getAllEntries(DBCACHE).next()
    rec2 = RecordedProgram((rec.chanid, rec.starttime), db=DBCACHE)

@test
@requires_becache
def test_recordedprogram_fromrecorded():
    rec = Recorded.getAllEntries(DBCACHE).next()
    rec2 = RecordedProgram.fromRecorded(rec)

test_recordedprogram.close()

test_oldrecorded = testgroup('OldRecorded')

@test
@requires_becache
def test_oldrecorded_getall():
    rec = list(OldRecorded.getAllEntries(DBCACHE))

@test
@requires_becache
def test_oldrecorded_getone():
    rec = OldRecorded.getAllEntries(DBCACHE).next()
    rec2 = OldRecorded((rec.chanid, rec.starttime), db=DBCACHE)

test_oldrecorded.close()

test_job = testgroup('Job')

@test
@requires_becache
def test_job_getall():
    job = list(Job.getAllEntries(DBCACHE))

@test
@requires_becache
def test_job_getone():
    try:
        job = Job.getAllEntries(DBCACHE).next()
    except StopIteration:
        raise Skipped
    job2 = Job(job.id)

test_job.close()

test_guide = testgroup('Guide')

@test
@requires_becache
def test_guide_getall():
    global guide
    guide = list(Guide.getAllEntries(DBCACHE))

@test
@requires_becache
def test_guide_getone():
    global guide
    if len(guide) == 0:
        raise Skipped
    guide2 = Guide((guide[0].chanid, guide[0].starttime), db=DBCACHE)

@test
@requires_becache
def test_guide_record():
    global guide
    for prog in guide:
        if (prog.starttime > datetime.now()+timedelta(1)) \
                and not prog.getRecStatus():
            break
    else:
        raise Skipped
    record = Record.fromGuide(prog)
    if len(record.getUpcoming()) == 0:
        raise Exception('new recording not scheduled')
    record.delete()

test_guide.close()

test_channel = testgroup('Channel')

@test
@requires_becache
def test_chan_getall():
    chan = list(Channel.getAllEntries(DBCACHE))

@test
@requires_becache
def test_chan_getone():
    try:
        chan = Channel.getAllEntries(DBCACHE).next()
    except StopIteration:
        raise Skipped
    chan2 = Channel(chan.chanid, db=DBCACHE)

test_channel.close()

test_video = testgroup('Video')

@test
@requires_dbcache
def test_video_getall():
    vids = list(Video.getAllEntries(DBCACHE))

@test
@requires_dbcache
def test_video_getone():
    try:
        vid = Video.getAllEntries(DBCACHE).next()
    except StopIteration:
        raise Skipped
    vid2 = Video(vid.intid, db=DBCACHE)

@test
@requires_dbcache
global vid
    vid = Video(db=DBCACHE)
    vid.title = 'test video'
    vid.subtitle = 'the first test'
    vid.episode = 1
    vid.season = 1
    vid.category = 'Test'
    vid.filename = 'testfile.avi'


test_video.close()

test_data_objects.close()

test_method_collections = testgroup('method collections')
test_mythbe = testgroup('MythBE')

@test
@requires_becache
def test_mythbe_getrecordings():
    be = MythBE(db=DBCACHE)
    recs = be.getRecordings()
    rec = be.getRecording(recs[0].chanid, int(mktime(recs[0].starttime.timetuple())))
    be.getPendingRecordings()
    be.getUpcomingRecordings()
    be.getScheduledRecordings()
    be.getConflictedRecordings()

test_mythbe.close()

test_mythdb = testgroup('MythDB')

@test
@requires_dbcache
def test_searchRecorded():
    global DBCACHE
    DBCACHE = MythDB(DBCACHE)
    db = DBCACHE
    list(db.searchRecorded())
    list(db.searchRecorded(commflagged=True))

@test
@requires_dbcache
def test_searchOldRecorded():
    db = DBCACHE
    list(db.searchOldRecorded())
    list(db.searchRecorded(commflagged=True))

@test
@requires_dbcache
def test_searchJobs():
    db = DBCACHE
    list(db.searchJobs())
    list(db.searchRecorded(commflagged=True))

@test
@requires_dbcache
def test_searchGuide():
    db = DBCACHE
    list(db.searchGuide())
    list(db.searchRecorded(commflagged=True))

@test
@requires_dbcache
def test_searchRecord():
    db = DBCACHE
    list(db.searchRecord())
    list(db.searchRecorded(commflagged=True))

test_mythdb.close()

test_mythxml = testgroup('MythXML')

@test
@requires_dbcache
def test_mythxml_hosts():
    global MYTHXML
    MYTHXML = MythXML(db=DBCACHE)
    xml = MYTHXML
    if len(xml.getHosts()) == 0:
        raise Exception('Empty hosts list')

@test
@requires_dbcache
def test_mythxml_keys():
    xml = MYTHXML
    xml = MythXML()
    if len(xml.getKeys()) == 0:
        raise Exception('Empty keys list')

@test
@requires_dbcache
def test_mythxml_getkey():
    xml = MYTHXML
    if int(xml.getSetting('DBSchemaVer')) != static.SCHEMA_VERSION:
        raise Exception('Incorrect schema version')

@test
@requires_dbcache
def test_mythxml_programs():
    xml = MYTHXML
    xml.getRecorded().next()

@test
@requires_dbcache
def test_mythxml_program():
    xml = MythXML()
    prog = xml.getRecorded().next()
    xml.getProgramDetails(prog.chanid, prog.starttime)

@test
@requires_dbcache
def test_mythxml_guide():
    xml = MYTHXML
    now = datetime.now()
    fut = datetime.now()+timedelta(1)
    chan = Guide.getAllEntries().next().chanid
    xml.getProgramGuide(now, fut, chan).next()

@test
@requires_dbcache
def test_mythxml_expiring():
    xml = MYTHXML
    xml.getExpiring()

test_mythxml.close()

test_mythvideo = testgroup('MythVideo')

@test
@requires_dbcache
def test_mythvideo_scan():
    mvid = MythVideo()
    mvid.scanStorageGroups(False)

@test
@requires_dbcache
def test_mythvideo_search():
    mvid = MythVideo()
    mvid.searchVideos().next()
    mvid.searchVideos(episode=1).next()

test_mythvideo.close()

test_method_collections.close()

test_collections.close()

#--------------------------------------
#

def finalprint():
    for fail in ERRORLIST:
        print '\n---------------------------------'
        print '%s Failed:' % fail[0]
        print fail[1]
    sys.exit(0)

def move_dbfile(moveback=False):
    cred = os.path.expanduser('~/.mythtv/config.xml')
    back = os.path.expanduser('~/.mythtv/config.xml.bak')
    if moveback:
        os.rename(back,cred)
    else:
        os.rename(cred, back)

pi = print_indent()
if __name__ == '__main__':
    try:
        runtests()
#    except: raise
    except: pass
    finalprint()

